/**
 * File: tree-map.h
 * ----------------
 * Defines a generic TreeMap, which implements to the
 * same interface that your Map class does, and even uses
 * the same data structure to back it.
 */

#pragma once

template <typename Key, typename Value>
class TreeMap {

public:
    /**
     * Constructor: TreeMap
     * --------------------
     * Constructs an empty TreeMap.
     */
    TreeMap();

    /**
     * Destructor: ~TreeMap
     * --------------------
     * Disposes of all resources currently held by the TreeMap.
     */
    ~TreeMap();
    
    /**
     * Method: isEmpty
     * ---------------
     * Returns true if and only if the TreeMap is empty.
     */
    bool isEmpty() const { return size() == 0; }

    /**
     * Method: size
     * ------------
     * Returns the number of key/value pairs stored in the TreeMap.
     */
    int size() const { return count; }
    
    /**
     * Method: containsKey
     * -------------------
     * Returns true if and only if the supplied key
     * is present in the TreeMap.
     */
    bool containsKey(const Key& key) const;

    /**
     * Method: put
     * -----------
     * Ensures that the supplied key/value pair is in the TreeMap.
     * Any value previously associated with the same key is overwritten.
     */
    void put(const Key& key, const Value& value);

    /**
     * Method: get
     * -----------
     * Returns a copy of the value attached to the provided key.
     * If the key isn't present, get returns a defaultly constructed
     * value (although the client should never rely on this behavior,
     * and instead should guard against such calls by first calling
     * containsKey).
     */
    Value get(const Key& key) const;

    /**
     * Method: operator[]
     * ------------------
     * Returns a reference to the value attached to the supplied
     * key.  If the key isn't present at the time of the call, then
     * the key is inserted, attached to a defaultly constructed
     * value, and a reference to that new value is returned.
     */
    Value& operator[](const Key& key);
    
private:
    struct node {
        Key key;
        Value value;
        node *left, *right;
    };
    
    int count;
    node *root;
    
    /**
     * Private Method: disposeTree
     * ---------------------------
     * Recursively disposes of everything reachable from the
     * supplied root node.
     */
    void disposeTree(node *root);

    /**
     * Private Method: findNode
     * ------------------------
     * Returns a reference to the address of the node
     * housing the supplied key.  If the key is missing,
     * findNode returns a reference to the pointer housing
     * a NULL that would have contained the address of the relevant
     * node had the key been present.
     */
    const node * const & findNode(const Key& key) const;

    /**
     * Private Method: ensureNodeExists
     * --------------------------------
     * If necessary, inserts a node on behalf of the supplied
     * key.  Whether originally present or freshly inserted,
     * ensureNodeExists returns the address of the relevant node.
     */
    node *ensureNodeExists(const Key& key);
};

#include "tree-map-impl.h"

